/*
 *  
 *  Copyright (c) 2004 Steve Slaven, All Rights Reserved.
 *  
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *  MA 02111-1307 USA
 *  
*/
package net.hoopajoo.cheeselab3;
import java.io.*;
import java.net.Socket;
import java.sql.*;
import java.util.HashMap;

//TODO: use channels here so we can freakin select
public class BrokerThread extends Thread {
    int max_search_results = 100;
    Socket sock = null;
    BufferedInputStream in = null;
    PrintStream out = null;
    boolean running = true;
    long last_syncid = -1;
    long last_tiles_syncid = -1;
    long last_world_syncid = -1;

    int last_sendid = 0;
    Connection db;
    String userid = null;
    CheesePlayerData player_data;
    private HashMap sentPlayers = new HashMap();

    // Reuse some prepared statements
    PreparedStatement statement_map = null;

    BrokerThread( Socket s ) {
        sock = s;
        try {
            s.setKeepAlive( true );
            in = new BufferedInputStream( sock.getInputStream() );
            out = new PrintStream( sock.getOutputStream(), true );
        } catch (IOException e) {}
    }

    public void run() {
        try {
            /*
            Class.forName( "org.postgresql.Driver" );
            db = DriverManager.getConnection( "jdbc:postgresql://localhost/cheeselab3", "postgres", "postgres" );
            */
            db = DriverManager.getConnection( "jdbc:derby:cheesedb" );
        } catch( Exception e ) {
            e.printStackTrace();
            return;
        }

        // Loop until we get a valid auth packet
        CheeseLab.dmesg( 1, "Accepting auth" );
        CheeseTx tx = CheeseTx.accept_tx( in );

        if( tx.msgtype.equals( "auth" ) ) {
            String authin = new String( tx.buffer );
            CheeseLab.dmesg( 2, "Parsing: " + authin );
            String[] auth = authin.split( ":" );
            String password = "";
            if( auth.length > 1 )
                password = auth[ 1 ];

            try {
                PreparedStatement s = db.prepareStatement( "select password from user_t where userid=?" );
                s.setString( 1, auth[ 0 ] );
                ResultSet rs = s.executeQuery();
                if( rs.next() ) {
                    if( password.equals( rs.getString( 1 ) ) ) {
                        CheeseLab.dmesg( 2, "Auth OK!" );
                        userid = auth[ 0 ];
                        send_tx( "auth", "OK" );
                    } else {
                        running = false; // exit
                        CheeseLab.dmesg( 2, "Auth failed" );
                        send_tx( "message", "Login failed" );
                    }
                } else {
                    running = false;
                    send_tx( "message", "Login failed" );
                }
                rs.close();
                s.close();
            } catch( Exception e ) {
                e.printStackTrace();
                CheeseLab.dmesg( 0, "Exception: " + e );
                send_tx( "message", "suck" );
                running = false;
            }
        } else {
            send_tx( "message", "Expected auth" );
            CheeseLab.dmesg( 1, "Did not send auth" );
            running = false;
        }

        // Go ahead and sync, this will force other brokers to try to print and hopefully
        // die if disconnected
        CheeseProxy.changed();

        CheeseLab.dmesg( 1, "Starting broker thread" );
        while( running ) {
            // player update
            if( CheeseProxy.get_syncid() > last_syncid ) {
                last_syncid = CheeseProxy.get_syncid();
                // send player update
                CheeseLab.dmesg( 2, "Players changed" );
                cmd_poll( null );
            }

            if( CheeseProxy.get_world_syncid() > last_world_syncid ) {
                last_world_syncid = CheeseProxy.get_world_syncid();
                // notify the client the world has changed, if they do not
                // get "world" tx's then they can aggressively optimize the display
                // better
                CheeseLab.dmesg( 2, "World changed" );
                send_tx( "world", "" );
            }

            if( CheeseProxy.get_tiles_syncid() > last_tiles_syncid ) {
                // Resend player tiles out
                last_tiles_syncid = CheeseProxy.get_tiles_syncid();

                File tiles = new File( "player_tiles" );
                File[] tilenames = tiles.listFiles();

                if( tilenames != null ) {
                    for( int i = 0; i < tilenames.length; i++ ) {
                        CheeseFile f = new CheeseFile( tilenames[ i ].getAbsolutePath() );
                        send_tx( "tile", CheeseDecoder.encode_object( f ) );
                    }
                }
            }


            // Check for incoming data
            try {
                if( in.available() > 0 ) {
                    // Handle TX
                    tx = CheeseTx.accept_tx( in );

                    // Get the latest player info for permissions checks
                    player_data = CheeseDecoder.get_player( userid, db );

                    if( tx.msgtype.equals( "poll" ) )
                        cmd_poll( tx );
                    else if( tx.msgtype.equals( "get" ) )
                        cmd_get( tx );
                    else if( tx.msgtype.equals( "map" ) )
                        cmd_map( tx );
                    else if( tx.msgtype.equals( "ping" ) )
                        cmd_ping( tx );
                    else if( tx.msgtype.equals( "save" ) && player_data.can_save() )
                        cmd_save( tx );
                    else if( tx.msgtype.equals( "move" ) && player_data.can_save() )
                        cmd_move( tx );
                    else if( tx.msgtype.equals( "merge" ) && player_data.can_save() )
                        cmd_merge( tx );
                    else if( tx.msgtype.equals( "delete" ) && player_data.can_save() )
                        cmd_delete( tx );
                    else if( tx.msgtype.equals( "search" ) )
                        cmd_search( tx );
                    else if( tx.msgtype.equals( "refresh" ) )
                        cmd_refresh( tx );
                    else if( tx.msgtype.equals( "tile" ) )
                        cmd_tile( tx );
                    else {
                        CheeseLab.dmesg( 1, "Unhandled msg: " + tx.msgtype );
                        send_tx( "message", "Unknown command or insufficient privileges for: " + tx.msgtype );
                    }
                }
            } catch( Exception e ) {
                CheeseLab.dmesg( 0, "Socket exception: " + e );
                e.printStackTrace();
                running = false;
            }

            try {
                // Loop around if we can
                if( in.available() < 1 ) {
                    // Don't sleep .25 seconds, since that is the heartbeat time in MUME
                    // we assume we did a little work and sleep a little less
                    sleep( 100 );

                }
            } catch( Exception e ) {}
        }

        // shut it down!
        CheeseLab.dmesg( 1, "Shutdown broker" );
        try {
            out.close();
        } catch( Exception e ) {
            e.printStackTrace();
        }
        try {
            out.close();
        } catch( Exception e ) {
            e.printStackTrace();
        }
        try {
            out.close();
        } catch( Exception e ) {
            e.printStackTrace();
        }
    }

    // Commands
    private void cmd_poll( CheeseTx tx ) {
        CheeseLab.dmesg( 2, "poll" );
        try {
            PreparedStatement s = db.prepareStatement( "select userid from user_t order by userid" );
            ResultSet rs = s.executeQuery();

            while( rs.next() ) {
                // Decode in to a player object
                String userid = rs.getString( "userid" );
                CheesePlayerData p = CheeseDecoder.get_player( userid, db );

                /* See if the player data matches the one we have previously
                 * sent to this client. If so, skip it. This will save us
                 * a lot of unnecessary traffic between server & client
                 */

                CheesePlayerData op = (CheesePlayerData)sentPlayers.get(userid);

                if( op == null || !op.equals(p) ) {
                    sentPlayers.put( userid, p );
                    CheeseLab.dmesg( 4, "Sending update for player: " + userid );
                    send_tx( "player", CheeseDecoder.encode_object( p ) );
                }
                else {
                    CheeseLab.dmesg( 4, "Player: " + userid + " status hasn't changed, skipping." );
                }
            }
            rs.close();
            s.close();
        } catch( Exception e ) {
            e.printStackTrace();
        }

    }

    private void cmd_get( CheeseTx tx ) {
        CheeseLab.dmesg( 2, "Get" );
        String line = new String( tx.buffer );
        String[] roomids = line.split( "," );

        String query = "select * from room_t where ";
        for( int i = 0; i < roomids.length; i++ ) {
            if( i > 0 )
                query += " OR ";

            query += "(roomid=? AND mtime>?)";
        }

        try {
            PreparedStatement s = db.prepareStatement( query );
            for( int i = 0; i < roomids.length; i++ ) {
                String[] split = roomids[ i ].split( ":" );
                s.setInt( i++ + 1, Integer.parseInt( split[ 0 ] ) );
                s.setLong( i + 1, Long.parseLong( split[ 1 ] ) );
                CheeseLab.dmesg( 4, "room: " + split[ 0 ] + " mtime>" + split[ 1 ] );
            }

            ResultSet rs = s.executeQuery();
            while( rs.next() ) {
                CheeseRoomData r = CheeseDecoder.sql_to_room( rs );
                send_tx( "room", CheeseDecoder.encode_object( r ) );
            }
            rs.close();
            s.close();
        } catch( Exception e ) {
            e.printStackTrace();
        }
    }

    /* This function still runs unacceptably slow, but I'm not sure why.
     * The slowness happens in the while( rs.next ) loop that takes about
     * 100ms to complete on my box, which drags down the mapper a lot.  In
     * the mean time we implement aggressive caching on the client side with 
     * remap commands happening in the background, but this is a problem for 
     * deleted room which will never leave the cache.... */
    private void cmd_map( CheeseTx tx ) {
        CheeseLab.dmesg( 2, "Map" );

        // Accept bounding box:
        // x,y,z xx,yy,zz
        String line = new String( tx.buffer );
        String[] triplets = line.split( " " );

        String[] from = triplets[ 0 ].split( "," );
        String[] to = triplets[ 1 ].split( "," );

        int from_x = Integer.parseInt( from[ 0 ] );
        int from_y = Integer.parseInt( from[ 1 ] );
        int from_z = Integer.parseInt( from[ 2 ] );

        int to_x = Integer.parseInt( to[ 0 ] );
        int to_y = Integer.parseInt( to[ 1 ] );
        int to_z = Integer.parseInt( to[ 2 ] );

        // Get rooms in range, send back data as roomid,x,y,z so it
        // can check it's cache and stuff
        try {
            java.util.Date start = new java.util.Date();
            if( statement_map == null ) {
                CheeseLab.dmesg( 1, "Preparing map statement" );
                statement_map = db.prepareStatement(
                                    "select roomid, x, y, z, mtime " +
                                    "from room_t where " +
                                    " x between ? and ? AND " +
                                    " y between ? and ? AND " +
                                    " z between ? and ?" );
            }

            statement_map.setInt( 1, from_x );
            statement_map.setInt( 2, to_x );
            statement_map.setInt( 3, from_y );
            statement_map.setInt( 4, to_y );
            statement_map.setInt( 5, from_z );
            statement_map.setInt( 6, to_z );

            ResultSet rs = statement_map.executeQuery();

            StringBuffer ret = new StringBuffer();

            /* Not sure if this is a good idea yet


            // Newer map clients using threaded rendering
            // can use this all zeroes line to sync their
            // rendered view with this map data view
            // while older clients will simply ignore this value
            ret.append( "0,0,0,0,0," + 
            	from_x +"," +
            	from_y +"," +
            	from_z +"," +
            	to_x +"," +
            	to_y +"," +
            	to_z +"\n" );
            */

            while( rs.next() ) {
                ret.append( rs.getString("roomid") );
                ret.append(",");
                ret.append( rs.getString("x") );
                ret.append(",");
                ret.append( rs.getString("y") );
                ret.append(",");
                ret.append( rs.getString("z") );
                ret.append(",");
                ret.append( rs.getString("mtime") );
                ret.append("\n");
            }

            // For profiling
            if( false ) {
                java.util.Date end = new java.util.Date();
                long rdiff = end.getTime() - start.getTime();
                double secs = (double)rdiff / 1000;
                CheeseLab.dmesg( 1, "Remap time: " + rdiff + "ms (" + secs + " secs)" );
            }

            // Simulate non-cached statements
            if( false ) {
                statement_map.close();
                statement_map = null;
            }

            rs.close();
            send_tx( "map", ret.toString() );
        } catch( Exception e ) {
            e.printStackTrace();
        }
    }
    private void cmd_ping( CheeseTx tx ) {
        CheeseLab.dmesg( 2, "Ping!" );
        last_sendid = CheeseTx.send_tx( out, last_sendid, 0, "ping", new String( tx.buffer ) );
    }
    private void cmd_save( CheeseTx tx ) {
        CheeseRoomData room = (CheeseRoomData)CheeseDecoder.decode_object( tx.buffer );
        CheeseDecoder.room_to_sql( room, db, userid );
        /* CheeseProxy.changed(); */
        CheeseProxy.world_changed();
    }
    private void cmd_move( CheeseTx tx ) {
        // Format is x,y,z offset, then roomid,roomid,roomid, all one line
        String line = new String( tx.buffer );
        String[] args = line.split( "," );

        // The number of roomid's is length - 3
        int roomids = args.length - 3;
        String update_sql = "update room_t" +
                            "	set userid=?, mtime=?," +
                            "	x=x+?, y=y+?, z=z+?" +
                            "	where roomid=0 ";
        for( int i = 0; i < roomids; i++ ) {
            //if( i > 0 )
            //	update_sql += " or ";

            update_sql += " or roomid=?";
        }

        CheeseLab.dmesg( 4, "SQL: " + update_sql );

        try {
            PreparedStatement s = db.prepareStatement( update_sql );

            int x = Integer.parseInt( args[ 0 ] ),
                    y = Integer.parseInt( args[ 1 ] ),
                        z = Integer.parseInt( args[ 2 ] );

            int col = 1;
            s.setString( col++, userid );
            s.setLong( col++, System.currentTimeMillis() );
            s.setInt( col++, x );
            s.setInt( col++, y );
            s.setInt( col++, z );

            for( int i = 3; i < args.length; i++ ) {
                int roomid = Integer.parseInt( args[ i ] );
                s.setInt( col++, roomid );
            }
            s.executeUpdate();
            s.close();
        } catch( Exception e ) {
            e.printStackTrace();
        }
        /* CheeseProxy.changed(); */
        CheeseProxy.world_changed();
    }

    private void cmd_delete( CheeseTx tx ) {
        String roomid_str = new String( tx.buffer );
        int roomid = Integer.parseInt( roomid_str );
        try {
            PreparedStatement update = db.prepareStatement( "delete from room_t where roomid=?" );
            update.setInt( 1, roomid );
            update.executeUpdate();

            String[] dirs = { "n", "s", "e", "w", "u", "d" };
            for( int i = 0; i < dirs.length; i++ ) {
                update = db.prepareStatement( "update room_t set exit_" + dirs[ i ] + "=-1, userid=?, mtime=? where exit_" + dirs[ i ] + "=?" );
                update.setInt( 3, roomid );
                update.setString( 1, userid );
                update.setLong( 2, System.currentTimeMillis() );
                update.executeUpdate();
            }
            update.close();
        } catch( Exception e ) {
            e.printStackTrace();
        }
        /* CheeseProxy.changed(); */
        CheeseProxy.world_changed();
    }

    private void cmd_search( CheeseTx tx ) {
        String search = new String( tx.buffer );

        // Strip beginning and ending % if passed to remain compatible with
        // older cheesemaps that assume we are doing one big "like" clause
        if( search.substring( 0, 1 ).equals( "%" ) ) {
            search = search.substring( 1, search.length() - 1 );
        }

        // Break up in to query parts, a query can be like:
        // word == search name
        // edesc=word == search edesc
        // fields that can be searched: name,notes,edesc
        // for multi-word searches we use the http url encoding of + to =
        // space
        String[] words = search.split( "\\s+" );
        String[] fields = new String[ words.length ];
        String[] values = new String[ words.length ];

        String[] valid_fields = { "name", "notes", "edesc" };

        for( int i = 0; i < words.length; i++ ) {
            boolean found = false;
            for( int w = 0; w < valid_fields.length && found == false; w++ ) {
                String keyword = valid_fields[ w ] + "=";
                CheeseLab.dmesg( 2, "Checking for '" + keyword + "' starting " + words[ i ] );
                if( words[ i ].length() > keyword.length() &&
                        words[ i ].substring( 0, keyword.length() ).equals( keyword ) ) {
                    fields[ i ] = valid_fields[ w ];
                    values[ i ] = words[ i ].substring( keyword.length() ).toLowerCase();
                    found = true;
                }
            }

            if( ! found ) {
                // Add it as name
                fields[ i ] = "name";
                values[ i ] = words[ i ].toLowerCase();
            }

            values[ i ] = values[ i ].replaceAll( "\\+", " " );
        }

        // Now convert the list of fields to a query
        StringBuffer query = new StringBuffer( "select roomid, name from room_t where " );
        for( int i = 0; i < fields.length; i++ ) {
            if( i > 0 ) {
                query.append( " AND " );
            }

            query.append( " lower(" + fields[ i ] + ") like ?" );
        }

        // And sorted by
        query.append( " order by name" );

        CheeseLab.dmesg( 2, "Generated search query is: " + query.toString() );
        try {
            send_tx( "progress", "Searching..." );
            PreparedStatement s = db.prepareStatement( query.toString() );

            // Populate our search vals
            for( int i = 0; i < values.length; i++ )
                s.setString( i + 1, "%" + values[ i ] + "%" );

            ResultSet rs = s.executeQuery();
            int cnt = 0;
            String results = "";
            while( rs.next() && cnt < max_search_results ) {
                cnt++;
                results += rs.getString( "roomid" ) + "," + rs.getString( "name" ) + "\n";
            }

            send_tx( "progress", "CLOSE" );
            send_tx( "search", results );
            if( rs.next() ) {
                send_tx( "message", "Search returned too many results, results truncated to " + max_search_results );
            }
            rs.close();
            s.close();

            if( cnt == 0 ) {
                send_tx( "message", "No rooms match search" );
            }
        } catch( Exception e ) {
            e.printStackTrace();
        }
    }

    private void cmd_refresh( CheeseTx tx ) {
        String last_mtime = new String( tx.buffer );

        boolean show_progress = true;
        if( last_mtime.indexOf( "noprogress" ) != -1 ) {
            show_progress = false;
            last_mtime = last_mtime.substring( 0, last_mtime.indexOf( "noprogress" ) - 1 );
        }


        if( show_progress )
            send_tx( "progress", "Selecting updates..." );

        try {
            long mtime = Long.parseLong( last_mtime );
            CheeseLab.dmesg( 2, "Refresh, mtime: " + last_mtime + "," + mtime );

            PreparedStatement s;
            ResultSet rs;
            int total = 0;

            if( show_progress ) {
                s = db.prepareStatement( "select count(*) as c from room_t where mtime>?" );
                s.setLong( 1, mtime );
                rs = s.executeQuery();
                rs.next();
                total = rs.getInt( "c" );
                rs.close();
                s.close();
            }

            s = db.prepareStatement( "select * from room_t where mtime>?" );
            s.setLong( 1, mtime );
            rs = s.executeQuery();
            int cnt = 0;
            while( rs.next() ) {
                CheeseRoomData r = CheeseDecoder.sql_to_room( rs );
                send_tx( "room", CheeseDecoder.encode_object( r ) );
                if( cnt++ % 100 == 0 && show_progress )
                    send_tx( "progress", "Accepted " + cnt + " of " + total + " rooms" );
            }
            rs.close();
            s.close();
        } catch( Exception e ) {
            e.printStackTrace();
        }
        if( show_progress )
            send_tx( "progress", "CLOSE" );

        send_tx( "refresh", "complete" );
    }

    private void cmd_merge( CheeseTx tx ) {
        String roomid_str = new String( tx.buffer );
        int roomid = Integer.parseInt( roomid_str );
        try {
            PreparedStatement s = db.prepareStatement( "select x,y,z from room_t where roomid=?" );
            s.setInt( 1, roomid );
            ResultSet rs = s.executeQuery();
            if( rs.next() ) {
                int x = rs.getInt( "x" ),
                        y = rs.getInt( "y" ),
                            z = rs.getInt( "z" );

                // Now go around n/s/e/w/u/d and connect any rooms that are 1 away
                // and have a zero exit coming the opposite way
                String[] from_fields = { "exit_n", "exit_s", "exit_e", "exit_w", "exit_u", "exit_d" };
                String[] to_fields = { "exit_s", "exit_n", "exit_w", "exit_e", "exit_d", "exit_u" };
                int[] offx = { 0, 0, 1, -1, 0, 0 };
                int[] offy = { -1, 1, 0, 0, 0, 0 };
                int[] offz = { 0, 0, 0, 0, 1, -1 };

                for( int i = 0; i < from_fields.length; i++ ) {
                    // First see if something is there
                    s = db.prepareStatement( "select roomid from room_t where x=? and y=? and z=? and " + to_fields[ i ] + "=-1" );
                    s.setInt( 1, x + offx[ i ] );
                    s.setInt( 2, y + offy[ i ] );
                    s.setInt( 3, z + offz[ i ] );
                    rs = s.executeQuery();
                    if( rs.next() ) {
                        // OK attach it to here from
                        int to_roomid = rs.getInt( "roomid" );
                        CheeseLab.dmesg( 3, "Gonna merge: " + to_roomid );

                        s = db.prepareStatement( "update room_t set " + from_fields[ i ] + "=?, userid=?, mtime=? where roomid=?" );
                        s.setInt( 1, to_roomid );
                        s.setString( 2, userid );
                        s.setLong( 3, System.currentTimeMillis() );
                        s.setInt( 4, roomid );
                        s.executeUpdate();


                        s = db.prepareStatement( "update room_t set " + to_fields[ i ] + "=?, userid=?, mtime=? where roomid=?" );
                        s.setInt( 1, roomid );
                        s.setString( 2, userid );
                        s.setLong( 3, System.currentTimeMillis() );
                        s.setInt( 4, to_roomid );
                        s.executeUpdate();
                    }
                }
            }
            rs.close();
            s.close();
        } catch( Exception e ) {
            e.printStackTrace();
        }
        /* CheeseProxy.changed(); */
        CheeseProxy.world_changed();
    }

    private void cmd_tile( CheeseTx tx ) {
        CheeseFile f = (CheeseFile)CheeseDecoder.decode_object( tx.buffer );
        f.save( "player_tiles" );

        // now all threads need to send out tiles again
        CheeseProxy.tiles_changed();
    }

    private void send_tx( String msgtype, String data ) {
        send_tx( msgtype, data.getBytes() );
    }
    private void send_tx( String msgtype, byte[] data ) {
        last_sendid = CheeseTx.send_tx( out, last_sendid, 0, msgtype, data );
        if( last_sendid == -1 ) {
            CheeseLab.dmesg( 1, "Failure sending, running is false" );
            running = false;
        }
    }
}
